/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2016-2017 Richard Hughes <richard@hughsie.com>
 *
 * Licensed under the GNU Lesser General Public License Version 2.1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifndef __FWUPD_ENUMS_PRIVATE_H
#define __FWUPD_ENUMS_PRIVATE_H

#define FWUPD_RESULT_KEY_APPSTREAM_ID		"AppstreamId"	/* s */
#define FWUPD_RESULT_KEY_CHECKSUM		"Checksum"	/* as */
#define FWUPD_RESULT_KEY_CREATED		"Created"	/* t */
#define FWUPD_RESULT_KEY_DESCRIPTION		"Description"	/* s */
#define FWUPD_RESULT_KEY_DEVICE_ID		"DeviceId"	/* s */
#define FWUPD_RESULT_KEY_FILENAME		"Filename"	/* s */
#define FWUPD_RESULT_KEY_FLAGS			"Flags"		/* t */
#define FWUPD_RESULT_KEY_FLASHES_LEFT		"FlashesLeft"	/* u */
#define FWUPD_RESULT_KEY_GUID			"Guid"		/* as */
#define FWUPD_RESULT_KEY_HOMEPAGE		"Homepage"	/* s */
#define FWUPD_RESULT_KEY_ICON			"Icon"		/* as */
#define FWUPD_RESULT_KEY_LICENSE		"License"	/* s */
#define FWUPD_RESULT_KEY_MODIFIED		"Modified"	/* t */
#define FWUPD_RESULT_KEY_METADATA		"Metadata"	/* a{ss} */
#define FWUPD_RESULT_KEY_NAME			"Name"		/* s */
#define FWUPD_RESULT_KEY_PLUGIN			"Plugin"	/* s */
#define FWUPD_RESULT_KEY_RELEASE		"Release"	/* a{sv} */
#define FWUPD_RESULT_KEY_REMOTE_ID		"RemoteId"	/* s */
#define FWUPD_RESULT_KEY_SIZE			"Size"		/* t */
#define FWUPD_RESULT_KEY_SUMMARY		"Summary"	/* s */
#define FWUPD_RESULT_KEY_TRUST_FLAGS		"TrustFlags"	/* t */
#define FWUPD_RESULT_KEY_UPDATE_ERROR		"UpdateError"	/* s */
#define FWUPD_RESULT_KEY_UPDATE_STATE		"UpdateState"	/* u */
#define FWUPD_RESULT_KEY_URI			"Uri"		/* s */
#define FWUPD_RESULT_KEY_VENDOR_ID		"VendorId"	/* s */
#define FWUPD_RESULT_KEY_VENDOR			"Vendor"	/* s */
#define FWUPD_RESULT_KEY_VENDOR			"Vendor"	/* s */
#define FWUPD_RESULT_KEY_VERSION_BOOTLOADER	"VersionBootloader"	/* s */
#define FWUPD_RESULT_KEY_VERSION_LOWEST		"VersionLowest"	/* s */
#define FWUPD_RESULT_KEY_VERSION		"Version"	/* s */

#endif /* __FWUPD_ENUMS_PRIVATE_H */
